@@ -12,6 +12,7 @@ import types
 import warnings
 from typing import (
     Any,
+    AsyncContextManager,
     AsyncGenerator,
     Callable,
     Coroutine,
@@ -27,15 +28,15 @@ from typing import (
     cast,
 )
 
-from ._types import Frame, Context, StackSlice
+from ._types import Stack, Frame, Context, StackSlice
 from ._code_dispatch import get_code
 from ._customization import (
     unwrap_stackitem,
     elaborate_frame,
     customize,
     unwrap_context,
+    unwrap_context_generator,
     elaborate_context,
-    fill_context,
     yields_frames,
 )
 from . import _extract
@@ -61,42 +62,72 @@ builtin_glue_pending: Dict[str, InstallGlueFn] = {}
 
 
 def builtin_glue(needs_module: str) -> Callable[[InstallGlueFn], InstallGlueFn]:
+    """Returns a decorator which marks the function it decorates as
+    providing glue for *needs_module*. The decorated function will be
+    called at most once, on the first attempt to extract a stack after
+    *needs_module* has been loaded, in order to set up customizations
+    that depend on that module.
+
+    Modules may also provide their own glue, by defining a top-level
+    function named ``_stackscope_install_glue_``. This function will
+    similarly be called at most once, on the first attempt to extract
+    a stack after the module defining it has been loaded. If both
+    module-provided and builtin glue exist for the same module, then
+    only the module-provided glue will run.
+    """
+
     def decorate(fn: InstallGlueFn) -> InstallGlueFn:
+        assert needs_module not in builtin_glue_pending
         if needs_module in sys.modules and "sphinx" not in sys.modules:
             fn()
         else:
-            assert needs_module not in builtin_glue_pending
             builtin_glue_pending[needs_module] = fn
         return fn
 
     return decorate
 
 
-def add_glue_as_needed() -> None:
-    for module_name in list(builtin_glue_pending):
-        if module_name in sys.modules:
+glue_lock = threading.Lock()
+
+
+def add_glue_as_needed(*, _sys_modules_len_cache: list[int] = [0]) -> None:
+    if len(sys.modules) == _sys_modules_len_cache[0]:
+        return
+    # Use a lock to avoid races between multiple threads trying to extract
+    # tracebacks simultaneously
+    with glue_lock:
+        module_names = tuple(sys.modules)
+        for module_name in module_names:
+            builtin_fn = builtin_glue_pending.pop(module_name, None)
             try:
-                install_fn = builtin_glue_pending.pop(module_name)
-            except KeyError:
-                # Probably two simultaneous calls occurring in
-                # different threads; the effect of glue is global
-                # so just skip it under the assumption that
-                # someone else got this one.
-                pass
-            else:
-                try:
-                    install_fn()
-                except Exception as exc:
-                    warnings.warn(
-                        "Failed to initialize glue for {}: {}. Some tracebacks may be "
-                        "presented less crisply or with missing information.".format(
-                            module_name,
-                            "".join(
-                                traceback.format_exception_only(type(exc), exc)
-                            ).strip(),
-                        ),
-                        RuntimeWarning,
-                    )
+                module_fn = sys.modules[module_name].__dict__.pop(
+                    "_stackscope_install_glue_", None
+                )
+            except Exception:  # module disappeared, doesn't have a dict, etc
+                module_fn = None
+            try:
+                # Prefer the module-supplied glue over our builtin version
+                # in case both are present
+                if module_fn is not None:
+                    module_fn()
+                elif builtin_fn is not None:
+                    builtin_fn()
+            except Exception as exc:
+                kind = (
+                    "module-provided" if module_fn is not None else "stackscope-builtin"
+                )
+                exc_str = "".join(
+                    traceback.format_exception_only(type(exc), exc)
+                ).strip()
+                warnings.warn(
+                    f"Failed to initialize {kind} glue for {module_name}: {exc_str}. "
+                    "Some tracebacks may be presented less crisply or with "
+                    "missing information.",
+                    RuntimeWarning,
+                )
+        # Only update the length cache if we visited every module (rather
+        # than bailing out with an exception)
+        _sys_modules_len_cache[0] = len(module_names)
 
 
 functools_singledispatch_wrapper = get_code(functools.singledispatch, "wrapper")
@@ -341,13 +372,37 @@ def glue_contextlib() -> None:
         # is currently exiting, since we'll see them later in the traceback
         # anyway
         if not context.is_exiting:
-            context.inner_stack = _extract.extract(mgr.gen)
+            context.inner_stack = _extract.extract_child(mgr.gen, for_task=False)
         if hasattr(mgr, "func"):
             context.description = format_funcall(mgr.func, mgr.args, mgr.kwds)
         else:
             # 3.7+ delete the func/args/etc attrs once entered
             context.description = f"{mgr.gen.__qualname__}(...)"
 
+    @unwrap_context.register(GCMBase)
+    def unwrap_generatorbased_contextmanager(mgr: Any, context: Context) -> Any:
+        mgr_code: types.CodeType
+        if hasattr(mgr.gen, "gi_code"):
+            mgr_code = mgr.gen.gi_code
+        elif hasattr(mgr.gen, "ag_code"):
+            mgr_code = mgr.gen.ag_code
+        else:  # pragma: no cover
+            return None
+        if mgr_code in unwrap_context_generator.registry:
+            if context.inner_stack is not None:
+                if context.inner_stack.frames:
+                    return unwrap_context_generator(
+                        context.inner_stack.frames[0], context
+                    )
+            else:
+                try:
+                    frame = _extract.extract_outermost(mgr.gen)
+                except RuntimeError:  # no frames
+                    pass
+                else:
+                    return unwrap_context_generator(frame, context)
+        return None
+
     @elaborate_context.register(ExitStackBase)
     def elaborate_exit_stack(stack: Any, context: Context) -> None:
         stackname = context.varname or "_"
@@ -409,7 +464,7 @@ def glue_contextlib() -> None:
                 varname=f"{stackname}[{idx}]",
                 start_line=context.start_line,
             )
-            fill_context(child_context)
+            _extract.fill_context(child_context)
             child_context.description = f"{tag}{stackname}.{method}({child_context.description or arg or '...'})"
             children.append(child_context)
 
@@ -477,7 +532,7 @@ def glue_async_generator() -> None:
         # is currently exiting, since we'll see them later in the traceback
         # anyway
         if not context.is_exiting:
-            context.inner_stack = _extract.extract(mgr._agen)
+            context.inner_stack = _extract.extract_child(mgr._agen, for_task=False)
         context.description = f"{mgr._func_name}(...)"
 
 
@@ -495,8 +550,20 @@ def glue_outcome() -> None:
     customize(outcome.acapture, hide=True)
 
 
+@builtin_glue("contextvars")
+def glue_contextvars() -> None:
+    import contextvars
+
+    # Don't show Context.run() in tracebacks even if it's implemented in
+    # Python (which is only true these days on pypy). On CPython it's
+    # implemented in C and will be hidden with or without this logic.
+    if isinstance(contextvars.Context.run, types.FunctionType):
+        customize(contextvars.Context.run, hide=True)
+
+
 @builtin_glue("trio")
 def glue_trio() -> None:
+    import threading
     import trio
 
     try:
@@ -518,6 +585,10 @@ def glue_trio() -> None:
         if hasattr(lowlevel, trap):  # pragma: no branch
             customize(getattr(lowlevel, trap), hide=True, prune=True)
 
+    @unwrap_stackitem.register(lowlevel.Task)
+    def unwrap_task(task: lowlevel.Task) -> Any:
+        return task.coro
+
     async def get_nursery_type() -> Type[Any]:
         return type(trio.open_nursery())
 
@@ -531,6 +602,185 @@ def glue_trio() -> None:
     @elaborate_context.register(nursery_manager_type)
     def elaborate_nursery(manager: Any, context: Context) -> None:
         context.obj = manager._nursery
+        assert isinstance(context.obj, trio.Nursery)
+        context.children = [
+            _extract.extract_child(child_task, for_task=True)
+            for child_task in context.obj.child_tasks
+        ]
+
+    @elaborate_frame.register(trio.to_thread.run_sync)
+    def elaborate_to_thread_run_sync(frame: Frame, next_inner: object) -> object:
+        thread_name = frame.pyframe.f_locals.get("thread_name")
+        worker_fn = frame.pyframe.f_locals.get("worker_fn")
+        sync_fn = frame.pyframe.f_locals.get("sync_fn")
+        if not (thread_name and worker_fn and sync_fn):  # pragma: no cover
+            # We're in the initial setup-y part, thread not running yet
+            return None
+
+        # Find the thread that's hosting the sync_fn
+        for thread in threading.enumerate():
+            if thread.name is thread_name:
+                break
+        else:  # pragma: no cover
+            # Thread isn't running yet
+            return None
+
+        # Find the actual frames where the sync_fn and its callees are running
+        inner_frame = sys._current_frames().get(thread.ident or 0)
+        previous: types.FrameType | None = None
+        current = inner_frame
+        while current is not None and current.f_code is not worker_fn.__code__:
+            previous = current
+            current = current.f_back
+        if current is None or previous is None:  # pragma: no cover
+            # We either didn't find the worker_fn, or it didn't have a callee.
+            # Thread isn't doing anything interesting yet.
+            return None
+
+        frame.hide = True
+        thread_stack = StackSlice(inner=inner_frame, outer=previous)
+        if (
+            isinstance(next_inner, Frame)
+            and next_inner.funcname == "wait_task_rescheduled"
+        ):
+            # Waiting for the thread to do something
+            return thread_stack
+        # Otherwise, the to_thread.run_sync task is most likely executing a
+        # reentrant from_thread.run_* call made by the sync_fn. Include the
+        # thread's stack but then switch back to the Trio task.
+        return (thread_stack, next_inner)
+
+    @elaborate_frame.register(trio.from_thread.run)
+    def elaborate_from_thread_run(frame: Frame, next_inner: object) -> object:
+        missing = object()
+        token_provided = frame.pyframe.f_locals.get("token_provided")
+        trio_token = frame.pyframe.f_locals.get("trio_token", missing)
+        if trio_token is missing:  # pragma: no cover
+            return None
+        if token_provided is None:  # pragma: no cover
+            # Trio v0.23.1 had token_provided; v0.24+ don't
+            token_provided = trio_token is not None
+        if not token_provided:
+            # No trio_token specified, so this is a reentrant call
+            # back into Trio from a to_thread.run_sync() function.
+            # The in-Trio portion will show up in the stack of
+            # to_thread.run_sync() so don't duplicate it here.
+            # Prune the plumbing (_send_message_to_trio, Queue.get(), etc)
+            # that's inward of here on the thread's stack.
+            frame.hide = True
+            return ()
+
+        # If a trio_token was specified, then this is a call that did not
+        # ultimately originate in Trio, so we should try to follow the link
+        # back to the Trio task if possible.
+        message = frame.pyframe.f_locals.get("message_to_trio")
+        if (
+            message is None
+            and isinstance(next_inner, Frame)
+            and next_inner.funcname == "_send_message_to_trio"
+        ):  # pragma: no cover  # depends on Trio minor version
+            message = next_inner.pyframe.f_locals.get("message_to_trio")
+        if message is not None:  # pragma: no branch
+            # Find the Trio runner that this token refers to
+            runner = None
+            try:
+                runner_tlocals = trio._core._run.GLOBAL_RUN_CONTEXT  # type: ignore
+            except AttributeError:  # pragma: no cover
+                return None
+            for ref in gc.get_referents(runner_tlocals):
+                if not isinstance(ref, dict):
+                    continue
+                for key, value in ref.items():
+                    if key == "runner" and value.trio_token is trio_token:
+                        # PyPy style: each thread's thread-locals dict is
+                        # directly referenced by the threading.local instance
+                        runner = value
+                        break
+                    if isinstance(value, dict) and (  # pragma: no branch
+                        runner := value.get("runner")
+                    ):
+                        # CPython style: each thread's thread-locals dict is
+                        # a value in one big dict keyed by weakrefs
+                        if runner.trio_token is trio_token:  # pragma: no branch
+                            break
+                else:  # pragma: no cover
+                    continue
+                break
+            else:  # pragma: no cover
+                return None
+
+            # Find the system task that matches this call
+            for task in runner.system_nursery.child_tasks:  # pragma: no branch
+                if task.context is message.context:  # pragma: no branch
+                    frame.hide = True
+                    return task.coro
+
+        return None  # pragma: no cover
+
+    if trio_threads := getattr(trio, "_threads", None):  # pragma: no branch
+        for clsname, fnname in (
+            ("Run", "run"),
+            ("Run", "run_system"),
+            ("Run", "unprotected_afn"),
+            ("RunSync", "run_sync"),
+            ("RunSync", "unprotected_fn"),
+        ):
+            cls = getattr(trio_threads, clsname, None)
+            fn = getattr(cls, fnname, None)
+            if fn is not None:  # pragma: no branch
+                customize(fn, hide=True)
+
+
+@builtin_glue("pytest_trio.plugin")
+def glue_pytest_trio() -> None:
+    import pytest_trio.plugin  # type: ignore
+    import trio
+
+    @elaborate_frame.register(pytest_trio.plugin.TrioFixture.run)
+    def elaborate_fixture(frame: Frame, next_inner: object) -> object:
+        # When a pytest-trio fixture is implemented using an async generator,
+        # any nurseries the fixture left open will be on the async generator's
+        # stack, which is not directly on the fixture task's stack. Detect
+        # this situation and fix it up.
+        if not isinstance(next_inner, Frame):  # pragma: no cover
+            return None
+        # Do we have the fixture generator yet? (Might not if we're still waiting
+        # for a depended-upon fixture to finish.)
+        gen = frame.pyframe.f_locals.get("func_value")
+        if gen is None:
+            return None
+        # Is it actually a generator? (Might not if this is a fixture that
+        # does some async setup and just returns.)
+        if not (hasattr(gen, "asend") or hasattr(gen, "__next__")):
+            return None
+        # Are we waiting for the test to finish? (Might not if we're still
+        # running the setup phase of the fixture.)
+        if next_inner.clsname == "Event" and next_inner.funcname == "wait":
+            # Yes -- show the fixture generator's stack in place of Event.wait().
+            frame.hide_line = True
+            return gen
+        return None
+
+    @unwrap_context_generator.register(pytest_trio.plugin.TrioFixture._fixture_manager)
+    def unwrap_fixture_manager(
+        frame: Frame, context: Context
+    ) -> None | AsyncContextManager[Any] | tuple[()]:
+        # Replace the internal _fixture_manager context (which is itself
+        # uninteresting) with the nursery that it contains (which might
+        # be hosting relevant portions of the test logic that we don't
+        # wish to hide). If that nursery has nothing in it, then hide it.
+        if frame.contexts and isinstance(
+            nursery := frame.contexts[0].obj, trio.Nursery
+        ):
+            if nursery.child_tasks:
+                return cast(AsyncContextManager[Any], nursery)
+            else:
+                return ()
+        else:  # pragma: no cover
+            # There are no checkpoints in _fixture_manager() except
+            # Nursery.__aexit__, so it is not generally possible to observe
+            # it without the nursery on its context stack.
+            return None
 
 
 @builtin_glue("greenlet")
@@ -547,7 +797,7 @@ def glue_greenlet() -> None:
             # otherwise a None frame means it's running
             if glet is not greenlet_getcurrent():
                 raise RuntimeError(
-                    "Can't dump the stack of a greenlet running " "in another thread"
+                    "Can't dump the stack of a greenlet running in another thread"
                 )
             # since it's running in this thread, its stack is our own
             inner_frame = get_true_caller()
@@ -630,5 +880,5 @@ def glue_greenback() -> None:
         return frame.pyframe.f_locals.get("coro")
 
     @unwrap_context.register(greenback.async_context)
-    def unwrap_greenback_async_context(manager: Any) -> Any:
+    def unwrap_greenback_async_context(manager: Any, context: Context) -> Any:
         return manager._cm
