@@ -71,7 +71,7 @@ def inspect_frame(frame: types.FrameType) -> FrameDetails:
     and evaluation stack for the currently executing or suspended
     frame *frame*.
 
-    There are three implementations of this function: one for CPython 3.7-3.10,
+    There are three implementations of this function: one for CPython 3.8-3.10,
     one for CPython 3.11+, and one for PyPy when using the "incminimark" garbage
     collector. The appropriate one will be chosen automatically.
     """
@@ -374,13 +374,7 @@ def currently_exiting_context(frame: types.FrameType) -> Optional[ExitingContext
         # the same as in the synchronous case
         offs -= 2
 
-    if sys.version_info < (3, 8):
-        # 3.7 and below: every exit call is done from a single WITH_CLEANUP_START
-        # location per 'with' block
-        if code[offs] == op["WITH_CLEANUP_START"]:
-            return ExitingContext(is_async=is_async, cleanup_offset=offs)
-        return None
-    elif sys.version_info < (3, 9):
+    if sys.version_info < (3, 9):
         # 3.8: they all use WITH_CLEANUP_START, but there might be multiple instances;
         # backtrack to the preceding POP_BLOCK
         if offs < 4 or code[offs] != op["WITH_CLEANUP_START"]:
@@ -799,9 +793,16 @@ def analyze_with_blocks(code: types.CodeType) -> Dict[int, Context]:
             # Allow for EXTENDED_ARG(s) before LOAD_CONST None
             while is_async and insns[idx + skip_insns - 5].opname == "EXTENDED_ARG":
                 skip_insns += 1
-            if is_async and sys.version_info >= (3, 12, 0, "alpha", 6):
-                # SEND stackeffect changed, resulting in an extra SWAP 2 + POP_TOP
-                skip_insns += 2
+            if is_async:
+                if sys.version_info >= (3, 12, 0, "beta", 1):
+                    # After 411b169281 there is an END_SEND bytecode after
+                    # the jump, to deal with changed SEND stackeffect more
+                    # efficiently than the below
+                    skip_insns += 1
+                elif sys.version_info >= (3, 12, 0, "alpha", 6):  # pragma: no cover
+                    # 160f2fe2b9 changed SEND stackeffect, resulting in
+                    # an extra SWAP 2 + POP_TOP
+                    skip_insns += 2
             store_to = describe_assignment_target(insns, idx + skip_insns)
             cleanup_offset = start_to_handler[insns[idx + skip_insns].offset]
             with_block_info[cleanup_offset] = Context(
